program lndcake;
uses Windows;


///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
/////////////////////////////// Consts \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
const
 VerData = 'LndCake v1.03';

 OpenLNDTitle: pchar = 'Open LND file';
 LNDFilter: packed array[1..30] of char = (
  'B', '&', 'W', ' ', 'l', 'a', 'n', 'd', ' ', 'f', 'i', 'l', 'e', 's', ' ', '(', '*', '.', 'l', 'n', 'd', ')', #0, '*', '.', 'L', 'N', 'D', #0,
  #0);
 SaveLNDTitle: pchar = 'Save LND file';
 SquishCountries: pchar = 'Do you want country (texture blending) data adjusted to match altitude changes?';
 SquishWithINI: pchar = 'Do you want to use custom squish rules?';
 INIFilter: packed array[1..25] of char = (
  'I', 'N', 'I', ' ', 'f', 'i', 'l', 'e', 's', ' ', '(', '*', '.', 'i', 'n', 'i', ')', #0, '*', '.', 'I', 'N', 'I', #0,
  #0);
 OpenINITitle: pchar = 'Open squish settings file';


///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
////////////////////////////// Typedefs \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
type
//TLndFile = imaginary record
// Header: TLndHeader;
// LoResTexture: array[1..Header.NumLoResTextures] of TLndLoResTex;
// Block: array[1..Header.NumBlocks] of TLndBlock;
// Country: array[1..Header.NumCountries] of TLndCountry;
// Material: array[1..Header.NumMaterials] of TLndMaterial;
// NoiseMap: array[0..255,0..255] of byte;
// BumpMap: array[0..255,0..255] of byte;
//end;

 TLndHeader = packed record
  NumBlocks: longint;
  BlockIndex: array[0..31,0..31] of byte;
  NumMaterials: longint;
  NumCountries: longint;
  BlockSize: longint;    // sizeof(TLndBlock)    =   2520
  MaterialSize: longint; // sizeof(TLndMaterial) = 131074
  CountrySize: longint;  // sizeof(TLndCountry)  =   3076
  NumLoResTextures: longword;
 end;

 TLndLoResTex = packed record
  Texture: pointer; // ignore
  Material: pointer; // ignore
  NumSubTextures: longint;
  ID: longint;
  Size: longint;
  // (Size - 4) bytes of DirectDraw texture crap
 end;

 TLndCell = packed record
  Color: array[1..4] of byte; // or something
  Altitude: byte;
  SaveColor: byte;
  Flags: array[1..2] of byte;
 end;
 TLndBlock = packed record
  Cell: array[0..16,0..16] of TLndCell;
  Index: longint;
  Dummy: array[1..204] of byte;
 end;

 TLndMapMaterial = packed record
  Index1: longword;
  Index2: longword;
  Coef1: longword;
 end;
 TLndCountry = packed record
  TerrainType: longword;
  MapMaterial: array[0..255] of TLndMapMaterial
 end;


///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
//////////////////////////////// Vars \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
var
 LND: HANDLE;
 IniFile: text;
 LNDOFN, INIOFN: OPENFILENAME;
 FNbuf, FNbuf2: array[1..320] of char;
 Header: TLndHeader;
 LoResTexture: TLndLoResTex;
 Block: TLndBlock;
 Country: TLndCountry;
 X, Y, Z, BytesRead, BytesWritten: longword;
 i, j: word;
 B, C: byte;
 SquishedAlt, SquishedCountry: array[0..255] of byte;
 IniSection, S: string;


///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
//////////////////////////////// Code \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
begin
 writeln(VerData);
 writeln;

 // Skip this if command line gives a filename?
 LNDOFN.lStructSize := sizeof(LNDOFN);
 LNDOFN.lpstrFilter := @LNDFilter;
 LNDOFN.lpstrFile := @FNbuf;
 LNDOFN.nMaxFile := sizeof(FNbuf);
 LNDOFN.lpstrTitle := OpenLNDTitle;
 LNDOFN.Flags := OFN_EXPLORER + OFN_FILEMUSTEXIST + OFN_HIDEREADONLY;
 if GetOpenFileName(@LNDOFN) then
  writeln('Loading ', FNbuf, '...')
 else // GetOpenFileName() failed (user probably cancelled).
  exit;

 FNbuf2 := FNbuf;
 LNDOFN.lpstrFile := @FNbuf2;
 LNDOFN.lpstrTitle := SaveLNDTitle;
 LNDOFN.Flags := OFN_EXPLORER + OFN_OVERWRITEPROMPT;
 if GetSaveFileName(@LNDOFN) then
  if not CopyFile(@FNbuf, @FNbuf2, FALSE) then
   begin
    writeln('Error ', GetLastError, ' preparing output file!');
    exit;
   end
  else // CopyFile() successful.
   writeln('Opening ', FNbuf2, '...')
 else // GetSaveFileName() failed (user probably cancelled).
  exit;

 LND := CreateFile(@FNbuf2, (GENERIC_READ + GENERIC_WRITE), FILE_SHARE_READ, nil, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
 if (LND = INVALID_HANDLE_VALUE) then
  begin
   writeln('Error ', GetLastError, ' opening output file!');
   exit;
  end;

 if not ReadFile(LND, Header, SizeOf(Header), BytesRead, nil) then
  begin
   writeln('Error ', GetLastError, ' reading LND header!');
   exit;
  end;

 writeln('Header info:');
 writeln(' NumBlocks        ', Header.NumBlocks);
 writeln(' NumMaterials     ', Header.NumMaterials);
 writeln(' NumCountries     ', Header.NumCountries);
 writeln(' BlockSize        ', Header.BlockSize);
 writeln(' MaterialSize     ', Header.MaterialSize);
 writeln(' CountrySize      ', Header.CountrySize);
 writeln(' NumLoResTextures ', Header.NumLoResTextures);
 writeln;

 for i := 0 to 255 do
  begin // Default squish settings
   case i of
    0..4: SquishedAlt[i] := i;
    5..7: SquishedAlt[i] := 4;
    8..10: SquishedAlt[i] := 5;
   else
    SquishedAlt[i] := i shr 1;
   end;
   case i of
    0..4: SquishedCountry[i] := i;
    5: SquishedCountry[i] := 8;
    6..127: SquishedCountry[i] := i shl 1;
   else
    SquishedCountry[i] := 255;
   end;
  end;

 if (MessageBox(0, SquishWithINI, OpenINITitle, MB_YESNO) = IDYES) then
  begin // Get custom squish settings
   FNbuf := '';
   INIOFN.lStructSize := sizeof(INIOFN);
   INIOFN.lpstrFilter := @INIFilter;
   INIOFN.lpstrFile := @FNbuf;
   INIOFN.nMaxFile := sizeof(FNbuf);
   INIOFN.lpstrTitle := OpenINITitle;
   INIOFN.Flags := OFN_EXPLORER + OFN_FILEMUSTEXIST + OFN_HIDEREADONLY;

   if GetOpenFileName(@INIOFN) then
    begin
     writeln('Loading ', FNbuf, '...');

     assign(IniFile, @FNbuf);
     reset(IniFile);
     while not eof(IniFile) do
      begin
       readln(IniFile, S);
       if (pos(';', S) > 0) then // Ignore comments
        SetLength(S, (pos(';', S) - 1));

       if (pos('[', S) > 0) and (pos(']', S) > pos('[', S)) then
        begin
         writeln;
         writeln(copy(S, pos('[', S), (1 + pos(']', S) - pos('[', S))));
         IniSection := upcase(copy(S, pos('[', S), (1 + pos(']', S) - pos('[', S))));
        end
       else if (pos('=', S) > 1) then
        begin
         val(copy(S, 1, (pos('=', S) - 1)), B, i);
         val(copy(S, (pos('=', S) + 1), (Length(S) - pos('=', S))), C, i);
         //writeln(B, '=', C);
         if IniSection = '[ALTITUDE]' then
          SquishedAlt[B] := C
         else if IniSection = '[COUNTRIES]' then
          SquishedCountry[B] := C;

        end;
      end;

     close(IniFile);
     writeln;
    end;
  end;

 for i := 1 to Header.NumLoResTextures do
  begin // skip low-res textures
   ReadFile(LND, LoResTexture, SizeOf(LoResTexture), BytesRead, nil);
   SetFilePointer(LND, (LoResTexture.Size - 4), nil, FILE_CURRENT);
  end;
 // ...OR jump to (filesize-(131072 + Header.NumCountries*Header.CountrySize + Header.NumMaterials*Header.MaterialSize + Header.NumBlocks*Header.BlockSize))

 writeln('Editing cell data:');
 for i := 2 to Header.NumBlocks do // why the hell...?
  begin
   Z := SetFilePointer(LND, 0, nil, FILE_CURRENT);
   ReadFile(LND, Block, SizeOf(Block), BytesRead, nil);
   write(' Block #', Block.Index, '...');
   for Y := 0 to 16 do
    for X := 0 to 16 do
     Block.Cell[Y,X].Altitude := SquishedAlt[Block.Cell[Y,X].Altitude];

   SetFilePointer(LND, Z, nil, FILE_BEGIN);
   WriteFile(LND, Block, SizeOf(Block), BytesWritten, nil);
   writeln(' done.');
  end;

 if (MessageBox(0, SquishCountries, SaveLNDTitle, MB_YESNO) = IDYES) then
  for i := 1 to Header.NumCountries do
   begin
    Z := SetFilePointer(LND, 0, nil, FILE_CURRENT);
    ReadFile(LND, Country, SizeOf(Country), BytesRead, nil);

    for j := 0 to 255 do
     Country.MapMaterial[j] := Country.MapMaterial[SquishedCountry[j]];

    SetFilePointer(LND, Z, nil, FILE_BEGIN);
    WriteFile(LND, Country, SizeOf(Country), BytesWritten, nil);
   end;

 writeln;
 writeln('Finished editing.');
 CloseHandle(LND);
end.
